/******************************************************************************
globusrun.c

Description:

CVS Information:

    $Source: /home/globdev/CVS/globus-1998_04_16/Globus/Miscellaneous/globusrun/globusrun.c,v $
    $Date: 1998/10/01 04:42:55 $
    $Revision: 1.40 $
    $Author: bester $
******************************************************************************/

/******************************************************************************
                             Include header files
******************************************************************************/
#include "globus_common.h"

#include <sys/types.h>
#include <sys/param.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>

#if !defined(PATH_MAX) && defined(MAXPATHLEN)
#   define PATH_MAX MAXPATHLEN 
#endif

#include "globus_nexus.h"
#include "globus_duroc_control.h"
#include "globus_gram_client.h"
#include "globus_gass_server_ez.h"
#include "globus_rsl.h"

#include "lber.h"
#include "ldap.h"

/******************************************************************************
                               Type definitions
******************************************************************************/
typedef struct globus_i_globusrun_gram_monitor_s
{
    globus_bool_t  done;
    globus_mutex_t mutex;
    globus_cond_t  cond;

    globus_bool_t  verbose;
    unsigned long  job_state;
} globus_i_globusrun_gram_monitor_t;

/*****************************************************************************
                          Module specific prototypes
*****************************************************************************/
/**** RSL Substitution of GASS redirection of stdout and stderr ****/
static int globus_l_globusrun_rsl_output_subst(globus_rsl_t *request,
					       char *globusrun_url_string);
/**** RSL Substitution of (dryrun=yes) for -dryrun flag ****/
static int globus_l_globusrun_rsl_dryrun_subst(globus_rsl_t *request);
/**** Run job using GRAM ****/
static int globus_l_globusrun_gramrun(char * request_string,
				      unsigned long options,
				      char *rm_contact);
/**** Run job using DUROC ****/
static int globus_l_globusrun_durocrun(char * request_ast,
				       unsigned long options,
				       int mpirun_version);
static int globus_l_globusrun_fault_callback(void *user_arg,
					      int fault_code);
static char **globus_l_globusrun_duroc_subjob_labels(char *request_string);

/**** LDAP search for resource manager name -> resource manager contact
      resolution *****/
static char *globus_l_globusrun_get_rm_contact(char *resource);
static char *globus_l_globusrun_get_contact_string(LDAP *ldap_server,
						   LDAPMessage *entry);
/**** globusrun job state file manipulations ****/
static char *globus_l_globusrun_create_globusrundir(void);
static char *globus_l_globusrun_create_job_file(void);
static void globus_l_globusrun_list_jobs(void);
static int  globus_l_globusrun_kill_job(int job_id,
					unsigned long options);
static void globus_l_globusrun_update_duroc_job_file(
    globus_duroc_control_t *control,
    char *job_contact,
    char *request_string,
    char *job_filename);

/**** Support for SIGINT handling ****/
static RETSIGTYPE globus_l_globusrun_sigint_handler(int dummy);
static int globus_l_globusrun_signal(int signum, RETSIGTYPE (*func)(int));
#if defined(BUILD_LITE)
    static int globus_l_globusrun_signal_wakeup(globus_bool_t blocking_allowed);
#   define globus_l_globusrun_remove_cancel_poll() globus_poll_remove(globus_l_globusrun_signal_wakeup)
#else
#   define globus_l_globusrun_remove_cancel_poll() 
#endif

/*****************************************************************************
                          Module specific variables
*****************************************************************************/
#define GLOBUSRUN_ARG_INTERACTIVE   	1
#define GLOBUSRUN_ARG_QUIET         	2
#define GLOBUSRUN_ARG_DRYRUN        	4
#define GLOBUSRUN_ARG_PARSE_ONLY    	8
#define GLOBUSRUN_ARG_AUTHENTICATE_ONLY 16
#define GLOBUSRUN_ARG_USE_GASS		32
#define GLOBUSRUN_ARG_ALLOW_READS      	64
#define GLOBUSRUN_ARG_ALLOW_WRITES 	128
#define GLOBUSRUN_ARG_IGNORE_CTRLC	256
#define GLOBUSRUN_ARG_BATCH             512

static globus_byte_t globus_l_globusrun_file_version=1;

#define GLOBUSRUN_USAGE \
"usage: globusrun [options] [RSL String]\n"\
"       globusrun -v\n"\
"Valid options are:\n"\
"    -i | -interactive \n"\
"           Run globusrun in interactive mode (multirequests only)\n"\
"    -f <rsl filename> | -file <rsl filename> \n"\
"           Read RSL from the local file <rsl filename>. The RSL can\n"\
"           be either a single job request, or a multirequest\n"\
"    -q | -quiet\n"\
"           Quiet mode (do not print diagnostic messages)\n"\
"    -o | -output-enable\n"\
"           Use the GASS Server library to redirect standout output\n"\
"           and standard error to globusrun. Implies -quiet\n"\
"    -s | -server\n"\
"           $(GLOBUSRUN_GASS_URL) can be used to access files local\n"\
"           to the submission machine via GASS. Implies -output-enable\n"\
"           and -quiet\n"\
"    -w | -write-allow\n"\
"           Enable the GASS Server library and allow writing to\n"\
"           GASS URLs. Implies -s and -q.\n"\
"    -r <resource manager name> | -r <resource manager contact> |\n"\
"        -resource-manager <resource manager name> |\n"\
"        -resource-manager <resource manager contact>\n"\
"           Submit the RSL job request to the specified resource manager.\n"\
"           A resource manager can be specified as a resource manager name\n"\
"           (<host>-<job manager type>) or a resource manager contact\n"\
"           <host>:<port>:<SSL DN>. This is a required argument when\n"\
"           submitting a single RSL request\n"\
"    -n | -no-interrupt\n"\
"           Cause SIGINT to terminate globusrun, while leaving the\n"\
"           submitted job to run to completion. By default the SIGINT\n"\
"           signal will be trapped and the job will be terminated\n"\
"    -l | -list-jobs\n"\
"           List disconnected globusrun jobs\n"\
"    -k <job number> | -kill <job number>\n"\
"           Kill a disconnected globusrun job\n"\
"    -b     Cause globusrun to terminate after the job is submited, without\n"\
"           waiting its completion. Usefull for batch jobs. to use -b is\n"\
"           equivalent to use -n and to send SIGINT to globusrun after\n"\
"           the job has started. This option can not be used together\n"\
"           with -s nor -i, and is also incompatible with multi request\n"\
"           jobs. Globusrun will write on its standard output a job ID.\n"\
"\n\nDIAGNOSTIC ARGUMENTS\n"\
"    -p | -parse-only\n"\
"           Parse and validate the RSL only. Do not submit the job to\n"\
"           a GRAM gatekeeper\n"\
"    -a | -authenticate-only\n"\
"           Submit a gatekeeper \"ping\" request only. Do not parse the\n"\
"           RSL or submit the job request. Requires the -r argument\n"\
"    -d | -dryrun\n"\
"           Submit the RSL to the job manager as a \"dryrun\" test\n"\
"           The request will be parsed and authenticated. The job manager\n"\
"           will execute all of the preliminary operations, and stop\n"\
"           just before the job request would be executed\n"

static globus_bool_t globus_l_globusrun_ctrlc = GLOBUS_FALSE;
static globus_bool_t globus_l_globusrun_ctrlc_handled = GLOBUS_FALSE;

/******************************************************************************
Function: main()

Description: 

Parameters: 

Returns: 
******************************************************************************/
int
main(int argc, char* argv[])
{
    char           *request_string = GLOBUS_NULL;
    char           *request_file = GLOBUS_NULL;
    globus_bool_t   no_more_options = GLOBUS_FALSE;
    globus_bool_t   usage_error = GLOBUS_FALSE;
    int             mpirun_version = 0;
    char           *rm_contact = GLOBUS_NULL;
    unsigned short  gass_port = 0;
    globus_rsl_t   *request_ast = GLOBUS_NULL;
    unsigned long   options = 0UL;
    int             err = GLOBUS_SUCCESS;
    int             i;
    globus_bool_t   ignore_ctrlc = GLOBUS_FALSE;
    
    err = globus_module_activate(GLOBUS_COMMON_MODULE);
    if ( err != GLOBUS_SUCCESS )
    {
	globus_libc_fprintf(stderr, "Error initializing globus\n");
	return 1;
    }
    for (i=1; i<argc; i++)
    {
	char *flag;
	
	flag=argv[i];

	if ( no_more_options 
	     || (flag[0]!='-') )
	{
	    /* last arg must be request string
	     * (after '--' flag or not a '-' flag) */
	    
	    if ( request_string != NULL )
	    {
		globus_libc_fprintf (stderr,
				     "%s: too many request "
				     "strings specified!\n",
				     argv[0]);
		usage_error = GLOBUS_TRUE;
	    }
	    else
	    {
		request_string = strdup (flag);
	    }
	}
	else
	{
	    /* option flag */
	    if ( flag[1]=='i' )
	    {
		options |= GLOBUSRUN_ARG_INTERACTIVE;
	    }
	    else if ( flag[1]=='f' )
	    {
		char *filename;
		
		filename = (char *) argv[++i];
		
		if ( request_file != NULL )
		{
		    globus_libc_fprintf (stderr,
			     "%s: too many request files specified!\n", 
			     argv[0]);
		    usage_error = GLOBUS_TRUE;
		}
		else
		{
		    request_file = strdup (filename);
		}
	    }
	    else if ( flag[1]=='q' )
	    {
		options |= GLOBUSRUN_ARG_QUIET;
	    }
	    else if ( flag[1]=='-' )
	    {
		no_more_options = 1;
	    }
	    else if ( flag[1]=='s' )
	    {
		options |= GLOBUSRUN_ARG_QUIET;
		options |= GLOBUSRUN_ARG_USE_GASS;
		options |= GLOBUSRUN_ARG_ALLOW_READS;
	    }
	    else if ( flag[1]=='o' )
	    {
		options |= GLOBUSRUN_ARG_QUIET;
		options |= GLOBUSRUN_ARG_USE_GASS;
	    }
	    else if ( flag[1]=='w')
	    {
		options |= GLOBUSRUN_ARG_QUIET;
		options |= GLOBUSRUN_ARG_USE_GASS;
		options |= GLOBUSRUN_ARG_ALLOW_READS;
		options |= GLOBUSRUN_ARG_ALLOW_WRITES;
	    }
	    else if ( flag[1] == 'p' )
	    {
		options |= GLOBUSRUN_ARG_PARSE_ONLY;
	    }
	    else if ( flag[1] == 'a' )
	    {
		options |= GLOBUSRUN_ARG_AUTHENTICATE_ONLY;
	    }
	    else if ( strcmp(flag,"-mpirun") == 0)
	    {
		if(++i == argc)
		{
		    usage_error = GLOBUS_TRUE;
		}
		else
		{
		    mpirun_version = atoi(argv[i]);
		}
	    }
	    else if ( flag[1]=='l' )
	    {
		globus_l_globusrun_list_jobs();
		globus_module_deactivate(GLOBUS_COMMON_MODULE);
		return 0;
	    }
	    else if ( flag[1]=='v' )
	    {
		  if(GLOBUS_RELEASE_NOT_BETA)
		  {
		      globus_libc_printf("Globus Version %d.%d.%d\n"
					 "GASS Protocol Version %d\n"
					 "GRAM Protocol Version %d\n",
					 globus_release_major(),
					 globus_release_minor(),
					 globus_release_patch(),
					 GLOBUS_GASS_PROTO_VERSION,
					 globus_gram_client_version());
		  }
		  else
		  {
		      globus_libc_printf("Globus Version %d.%d.%db%d\n"
					 "GASS Protocol Version %d\n"
					 "GRAM Protocol Version %d\n",
					 globus_release_major(),
					 globus_release_minor(),
					 globus_release_patch(),
					 globus_release_beta(),
					 GLOBUS_GASS_PROTO_VERSION,
					 globus_gram_client_version());
		      
		  }
		  globus_module_deactivate(GLOBUS_COMMON_MODULE);
		  return 0;
	    }
	    else if ( flag[1]=='k' )
	    {
		if(++i == argc)
		{
		    usage_error = GLOBUS_TRUE;
		}
		else
		{
		    if(atoi(argv[i]) == 0)
		    {
			globus_libc_printf("%s: Invalid job id %s\n",
					   argv[0], argv[i]);
			usage_error = GLOBUS_TRUE;
		    }
		    else
		    {
			globus_l_globusrun_kill_job(atoi(argv[i]), options);
			return 0;
		    }
		}
	    }
	    else if ( flag[1]=='r' )
	    {
		if(++i == argc)
		{
		    usage_error = GLOBUS_TRUE;
		}
		else
		{
		    rm_contact = globus_l_globusrun_get_rm_contact(argv[i]);
		    if(rm_contact == GLOBUS_NULL)
		    {
			globus_libc_fprintf(stderr,
					    "%s: Error resolving resource "
					    "manager contact %s\n",
					    argv[0],
					    argv[i]);
			return 0;
		    }
		}
	    }
	    else if ( flag[1] == 'd' )
	    {
		options |= GLOBUSRUN_ARG_DRYRUN;
	    }
	    else if (flag[1] == 'n')
	    {
		options |= GLOBUSRUN_ARG_IGNORE_CTRLC;
		ignore_ctrlc = GLOBUS_TRUE;
	    }
	    else if (flag[1] == 'b')
	    {
		options |=  GLOBUSRUN_ARG_BATCH;
	    }
	    else if (flag[1] == 'h' )
	    {
	        globus_libc_printf (GLOBUSRUN_USAGE);
		return 0;
	    }
	    else
	    {
		/* unrecognized flag */
		globus_libc_fprintf (stderr, "%s: unrecognized flag %s!\n",
			 argv[0], flag);
		usage_error = GLOBUS_TRUE;
	    }
	}
    }

    if ( (options & GLOBUSRUN_ARG_BATCH) &&
	 (options & GLOBUSRUN_ARG_USE_GASS) )
    {
	globus_libc_fprintf (
	    stderr, "%s: Option -s and -b are exclusive!\n",
	    argv[0]);
	usage_error = GLOBUS_TRUE;
    }

    if ( (options & GLOBUSRUN_ARG_BATCH) &&
	 (options & GLOBUSRUN_ARG_INTERACTIVE) )
    {
	globus_libc_fprintf (
	    stderr, "%s: Option -i and -b are exclusive!\n",
	    argv[0]);
	usage_error = GLOBUS_TRUE;
    }
    
    if(options & GLOBUSRUN_ARG_AUTHENTICATE_ONLY &&
       !usage_error)
    {
	err = globus_module_activate(GLOBUS_GRAM_CLIENT_MODULE);
	if(err != GLOBUS_SUCCESS)
	{
	    globus_libc_fprintf(stderr,
				"\n%s: Error initializing GRAM: %s\n",
				argv[0],
				globus_gram_client_error_string(err));
	    return 1;
	}
	err = globus_gram_client_ping(rm_contact);
	if(err == GLOBUS_SUCCESS)
	{
	    globus_libc_fprintf(stderr,
				"\nGRAM Authentication test successful\n");
	    return 0;
	}
	else
	{
	    globus_libc_fprintf(stderr,
				"\nGRAM Authentication test failure: %s\n",
				globus_gram_client_error_string(err));
	    return 1;
	}
	
    }
    if ( (request_string!=NULL)
	 && (request_file!=NULL) )
    {
	globus_libc_fprintf (stderr,
			     "%s: cannot specify both request "
			     "string and request file!\n",
			     argv[0]);
	usage_error = GLOBUS_TRUE;
    }
    
    if ( request_file != NULL)
    {
	int fd;
	
	fd = globus_libc_open (request_file, O_RDONLY, 0600);
	if ( fd >= 0 )
	{
	    int i;
	    char c;
	    off_t len = 0;
	    

	    len = lseek(fd, 0, SEEK_END);
	    lseek(fd, 0, SEEK_SET);
	    
	    request_string = (char *) malloc (sizeof (char)
					      * (len + 1));
	    i=0;
	    
	    while ( (i<=len)  && read(fd, &c, 1) > 0)
	    {
		request_string[i] = c;
		i++;
	    }
	    request_string[i] = '\0';
	    
	    free (request_file);
	    request_file = NULL;
	    
	    globus_libc_close (fd);
	}
	else
	{
	    globus_libc_fprintf (stderr,
				 "%s: cannot open file %s!\n",
				 argv[0],
				 request_file);
	    usage_error = 1;
	}
    }
    
    if ( (request_string==NULL)
	 && (request_file==NULL))
    {
	globus_libc_fprintf (stderr,
			     "%s: must specify a request "
			     "string or request file!\n",
			     argv[0]);
	usage_error = 1;
    }

    if(!usage_error)
    {
	request_ast = globus_rsl_parse(request_string);
	if(request_ast == GLOBUS_NULL)
	{
	    globus_libc_fprintf(stderr,
				"%s Error: Cannot Parse RSL %s\n",
				argv[0], request_string);
	}

	
	if(!globus_rsl_is_boolean(request_ast))
	{
	    globus_libc_fprintf(stderr,
				"%s Error: Bad RSL\n",
				argv[0]);
	    err=-2;
	    return err;
	}
	if(options & GLOBUSRUN_ARG_PARSE_ONLY)
	{
	    globus_libc_printf("RSL Parsed Successfully...\n");
	    return 0;
	}

	if((!globus_rsl_is_boolean_multi(request_ast)) &&
	   rm_contact == GLOBUS_NULL)
	{
	    usage_error=1;
	}
    }
    
    if ( usage_error )
    {
	globus_libc_fprintf (stderr, 
			     GLOBUSRUN_USAGE);
	return -1;
    }
    
    /* intialize and start nexus */
    err = globus_module_activate(GLOBUS_NEXUS_MODULE);
    if ( err != GLOBUS_SUCCESS )
    {
	globus_libc_fprintf (stderr, "Error initializing nexus\n");
	return -1;
    }
    
    globus_nexus_enable_fault_tolerance(globus_l_globusrun_fault_callback,
					NULL);
    
    
    if (options & GLOBUSRUN_ARG_USE_GASS)
    {
	/* transform the RSL to send "free" output streams to our
	 * stdout/stderr via gass, rather than to /dev/null */
	char *gass_server_url = NULL;
	unsigned long server_ez_opts=0UL;
	char *url_relation_string;
	char *relation_format="rsl_substitution=(GLOBUSRUN_GASS_URL %s)";
	    
	server_ez_opts |=
	    GLOBUS_GASS_SERVER_EZ_LINE_BUFFER
	    | GLOBUS_GASS_SERVER_EZ_STDOUT_ENABLE
	    | GLOBUS_GASS_SERVER_EZ_STDERR_ENABLE ;
	
	if(options & GLOBUSRUN_ARG_ALLOW_READS)
	{
	    server_ez_opts |= 
		GLOBUS_GASS_SERVER_EZ_READ_ENABLE;
	}
	if(options & GLOBUSRUN_ARG_ALLOW_WRITES)
	{
	    server_ez_opts |= 
		GLOBUS_GASS_SERVER_EZ_WRITE_ENABLE;
	}

	err = globus_gass_server_ez_init(&gass_port, 
					 &gass_server_url,
					 server_ez_opts,
					 NULL);
	
	if((err != GLOBUS_SUCCESS) ||
	   gass_server_url == GLOBUS_NULL)
	{
	    globus_libc_fprintf(stderr,
				"%s Error: initializing GASS (%d)\n",
				argv[0],
				err);
	    goto hard_exit;
	}
	
	url_relation_string = globus_malloc(strlen(relation_format) +
					    strlen(gass_server_url));

	if(url_relation_string == GLOBUS_NULL)
	{
	    err=-3;
	    
	    globus_libc_fprintf(stderr,
				"%s Error: Malloc failed\n",
				argv[0]);
	    goto hard_exit;
	}
	
	globus_libc_lock();
	sprintf(url_relation_string,
		relation_format,
		gass_server_url);
	globus_libc_unlock();
	
	/* replace stdout and stderr in request */
	globus_l_globusrun_rsl_output_subst(request_ast,
					    url_relation_string);


	globus_free (gass_server_url);
	globus_free (url_relation_string);
	
	gass_server_url = NULL;
    }
    if(options & GLOBUSRUN_ARG_DRYRUN)
    {
	globus_l_globusrun_rsl_dryrun_subst(request_ast);
    }
    if(request_string)
    {
	globus_free(request_string);
    }
    request_string = globus_rsl_unparse(request_ast);
    
    if(globus_rsl_is_boolean_multi(request_ast))
    {
	char *req;

	if ( options & GLOBUSRUN_ARG_BATCH )
	{
	    globus_libc_fprintf(
		stderr,
		"%sBatch mode (-b) not supported for mulirequest\n",
		argv[0]);
	    usage_error = GLOBUS_TRUE;
	    goto hard_exit;
	}
	    

	if(rm_contact != GLOBUS_NULL)
	{
	    globus_libc_fprintf(stderr,
				"%s warning: ignoring "
				"resource manager contact for mulirequest\n",
				argv[0]);
	}
	req = globus_rsl_unparse(request_ast);
	
	err = globus_l_globusrun_durocrun(req,
					  options,
					  mpirun_version);
	globus_free(req);
    }
    else
    {
	char *req;

	req = globus_rsl_unparse(request_ast);
	
	err = globus_l_globusrun_gramrun(req,
					 options,
					 rm_contact);
	globus_free(req);
    }    
    
hard_exit:
    globus_l_globusrun_signal(SIGINT, SIG_DFL);
    
    if ( (options & GLOBUSRUN_ARG_USE_GASS) &&
	 (gass_port != 0U))
    {
	globus_gass_server_ez_shutdown (gass_port);
    }
    
    globus_module_deactivate(GLOBUS_NEXUS_MODULE);
    
    globus_free (request_string);

    globus_module_deactivate(GLOBUS_COMMON_MODULE);
    
    return err;
} /* main() */

/******************************************************************************
Function: globus_l_globusrun_new_subst()

Description: 

Parameters: 

Returns: 
******************************************************************************/
static globus_rsl_t *
globus_l_globusrun_new_subst(char *var)
{
    globus_rsl_t *relation;
    char *relation_format = "%s=$(GLOBUSRUN_GASS_URL)/dev/%s";
    char *val;

    val = globus_malloc((2 * strlen(var)) +
			strlen(relation_format));

    globus_libc_lock();
    sprintf(val, relation_format, var, var);
    globus_libc_unlock();

    relation = globus_rsl_parse(val);

    globus_free(val);

    return relation;
} /* globus_l_globusrun_new_subst() */

/******************************************************************************
Function: globus_l_globusrun_var_exists()

Description: 

Parameters: 

Returns: 
******************************************************************************/
static globus_bool_t
globus_l_globusrun_var_exists(globus_rsl_t *subrequest,
			      char *val)
{
    if(globus_rsl_is_relation_eq(subrequest))
    {
	if(strcmp(globus_rsl_relation_get_attribute(subrequest), val) == 0)
	{
	    return GLOBUS_TRUE;
	}
    }
    return GLOBUS_FALSE;
} /* globus_l_globusrun_var_exists() */

/******************************************************************************
Function: globus_l_globusrun_rsl_dryrun_subst()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static int
globus_l_globusrun_rsl_dryrun_subst(globus_rsl_t *request)
{
    int level = 0;
    
    /* for multirequests */
    if(globus_rsl_is_boolean_multi(request))
    {
	globus_list_t *lists;

	lists = (globus_list_t *) globus_rsl_boolean_get_operand_list(request);

	while(!globus_list_empty(lists))
	{
	    level += globus_l_globusrun_rsl_dryrun_subst(
		globus_list_first(lists));
	    lists=globus_list_rest(lists);
	}
    }
    else if(globus_rsl_is_boolean(request))
    {
	globus_list_t *l;
	globus_bool_t dryrun_exists = GLOBUS_FALSE;
	
	l = globus_rsl_boolean_get_operand_list(request);
	while(!globus_list_empty(l))
	{

	    if(globus_rsl_is_boolean(globus_list_first(l)))
	    {
		level += globus_l_globusrun_rsl_dryrun_subst(
		    globus_list_first(l));
	    }

	    if(globus_l_globusrun_var_exists(globus_list_first(l),
						    "dryrun"))
	    {
		dryrun_exists = GLOBUS_TRUE;
	    }
	    
	    l = globus_list_rest(l);
	}
	
	if(level == 0)
	{
	    if(!dryrun_exists)
	    {
		globus_list_insert(
		    globus_rsl_boolean_get_operand_list_ref(request),
		    globus_rsl_parse("dryrun = yes"));
							    
	    }
	    level++;
	}
    }
    return level;
} /* globus_l_globusrun_rsl_dryrun_subst() */

/******************************************************************************
Function: globus_l_globusrun_rsl_output_subst()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static int
globus_l_globusrun_rsl_output_subst(globus_rsl_t *request,
				    char *globusrun_gass_url_string)
{
    int level = 0;
    
    /* for multirequests */
    if(globus_rsl_is_boolean_multi(request))
    {
	globus_list_t *lists;

	lists = (globus_list_t *) globus_rsl_boolean_get_operand_list(request);

	while(!globus_list_empty(lists))
	{
	    level += globus_l_globusrun_rsl_output_subst(
		globus_list_first(lists),
		globusrun_gass_url_string);
	    lists=globus_list_rest(lists);
	}
    }
    else if(globus_rsl_is_boolean(request))
    {
	globus_list_t *l;
	globus_bool_t stdout_exists = GLOBUS_FALSE;
	globus_bool_t stderr_exists = GLOBUS_FALSE;
	
	l = globus_rsl_boolean_get_operand_list(request);
	while(!globus_list_empty(l))
	{

	    if(globus_rsl_is_boolean(globus_list_first(l)))
	    {
		level += globus_l_globusrun_rsl_output_subst(
		    globus_list_first(l),
		    globusrun_gass_url_string);
	    }

	    if(globus_l_globusrun_var_exists(globus_list_first(l),
						    "stdout"))
	    {
		stdout_exists=GLOBUS_TRUE;
	    }
	    if(globus_l_globusrun_var_exists(globus_list_first(l),
						    "stderr"))
	    {
		stderr_exists=GLOBUS_TRUE;
	    }
	    
	    l = globus_list_rest(l);
	}
	
	if(level == 0)
	{
	    if(!stdout_exists)
	    {
		globus_list_insert(
		    globus_rsl_boolean_get_operand_list_ref(request),
		    globus_l_globusrun_new_subst("stdout"));
	    }
	    if(!stderr_exists)
	    {
		globus_list_insert(
		    globus_rsl_boolean_get_operand_list_ref(request),
		    globus_l_globusrun_new_subst("stderr"));
	    }
	    globus_list_insert(
		globus_rsl_boolean_get_operand_list_ref(request),
		globus_rsl_parse(strdup(globusrun_gass_url_string)));
	    level++;
	}
    }
    return level;
} /* globus_l_globusrun_rsl_output_subst() */

/******************************************************************************
Function: globus_l_globusrun_gram_callback_func()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static void
globus_l_globusrun_gram_callback_func(void *user_arg,
				      char *job_contact,
				      int state,
				      int errorcode)
{
    globus_i_globusrun_gram_monitor_t *monitor;

    monitor = (globus_i_globusrun_gram_monitor_t *) user_arg;

    globus_mutex_lock(&monitor->mutex);
    
    monitor->job_state = state;
    
    switch(state)
    {
    case GLOBUS_GRAM_CLIENT_JOB_STATE_PENDING:
	if(monitor->verbose)
	{
	    globus_libc_printf("GLOBUS_GRAM_CLIENT_JOB_STATE_PENDING\n");
	}
	break;
    case GLOBUS_GRAM_CLIENT_JOB_STATE_ACTIVE:
	if(monitor->verbose)
	{
	    globus_libc_printf("GLOBUS_GRAM_CLIENT_JOB_STATE_ACTIVE\n");
	}
	break;
    case GLOBUS_GRAM_CLIENT_JOB_STATE_FAILED:
	if(monitor->verbose)
	{
	    globus_libc_printf("GLOBUS_GRAM_CLIENT_JOB_STATE_FAILED\n");
	}
        monitor->done = GLOBUS_TRUE;
	break;
    case GLOBUS_GRAM_CLIENT_JOB_STATE_DONE:
	if(monitor->verbose)
	{
	    globus_libc_printf("GLOBUS_GRAM_CLIENT_JOB_STATE_DONE\n");
	}
        monitor->done = GLOBUS_TRUE;
	break;
    }

    globus_cond_signal(&monitor->cond);
    globus_mutex_unlock(&monitor->mutex);
} /* globus_l_globusrun_gram_callback_func() */

/******************************************************************************
Function: globus_l_globusrun_gramrun()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static int
globus_l_globusrun_gramrun(char * request_string,
			   unsigned long options,
			   char *rm_contact)
{
    char *callback_contact = GLOBUS_NULL;
    char *job_contact = GLOBUS_NULL;
    char *jobfilename = GLOBUS_NULL;
    FILE *fp;
    globus_i_globusrun_gram_monitor_t monitor;
    int err;
    globus_bool_t verbose = !(options & GLOBUSRUN_ARG_QUIET);

    /* trap SIGINTs */
    if(!(options & GLOBUSRUN_ARG_IGNORE_CTRLC))
    {
	globus_l_globusrun_signal(SIGINT,
                                  globus_l_globusrun_sigint_handler);
#       if defined(BUILD_LITE)
	{
	    globus_poll_add(globus_l_globusrun_signal_wakeup, 1);
	}
#       endif
    }

    monitor.done = GLOBUS_FALSE;
    monitor.verbose=verbose;
    globus_mutex_init(&monitor.mutex, GLOBUS_NULL);
    globus_cond_init(&monitor.cond, GLOBUS_NULL);

    err = globus_module_activate(GLOBUS_GRAM_CLIENT_MODULE);
    if(err != GLOBUS_SUCCESS)
    {
	globus_libc_fprintf(stderr,
			    "Error initializing GRAM client: %d - %s\n",
			    err,
			    globus_gram_client_error_string(err));
	goto hard_exit;
    }

    if(options & GLOBUSRUN_ARG_IGNORE_CTRLC)
    {
        globus_l_globusrun_signal(SIGINT,
                                  SIG_DFL);
#       if defined(BUILD_LITE)
	{
	    globus_poll_remove(globus_l_globusrun_signal_wakeup);
	}
#       endif
    }

    if(!(options & GLOBUSRUN_ARG_DRYRUN))
    {
	err = globus_gram_client_callback_allow(
	    globus_l_globusrun_gram_callback_func,
	    (void *) &monitor,
	    &callback_contact);

	if(err != GLOBUS_SUCCESS)
	{
	    globus_libc_fprintf(stderr,
				"Initializing GRAM Callback failed because %s (errorcode %d)\n",
				globus_gram_client_error_string(err),
				err);

	    goto hard_exit;
	}
	else if(verbose)
	{
	    globus_libc_printf("globus_gram_client_callback_allow "
			       "successful\n");
	}
    }
    
    err = globus_gram_client_job_request(rm_contact,
					 request_string,
					 GLOBUS_GRAM_CLIENT_JOB_STATE_ALL,
					 callback_contact,
					 &job_contact);

    if(err != GLOBUS_SUCCESS)
    {
	if(callback_contact)
	{
	    globus_gram_client_callback_disallow(callback_contact);
	    globus_free(callback_contact);
	}
	
	if((err == GLOBUS_GRAM_CLIENT_ERROR_DRYRUN) &&
	   (options & GLOBUSRUN_ARG_DRYRUN))
	{
	    if(verbose)
	    {
		globus_libc_fprintf(stderr,
				    "Dryrun successful\n");
	    }
	    err=0;
	}
	else
	{
	    globus_libc_fprintf(stderr,
				"GRAM Job submission failed because %s (error code %d)\n",
				globus_gram_client_error_string(err),
				err);
				
	}
	goto hard_exit;
    }
    else if(verbose)
    {
	globus_libc_printf("GRAM Job submission successful\n");
    }

    jobfilename = globus_l_globusrun_create_job_file();
    fp = fopen(jobfilename, "w");
    globus_libc_fprintf(fp,
			"GR%c:%d:%s%d:%s",
			globus_l_globusrun_file_version,
			(int) strlen(request_string),
			request_string,
			(int) strlen(job_contact),
			job_contact);
    fclose(fp);

     if  (options & GLOBUSRUN_ARG_BATCH)
     {
	 char * job_id_pt=jobfilename;
	 char * index;

	 if ( (index=strrchr(job_id_pt,'/')) != GLOBUS_NULL)
	 {
	     job_id_pt=index+1;
	 }
	 globus_libc_printf("%s\n",job_id_pt);
	 goto hard_exit;
     }
    globus_mutex_lock(&monitor.mutex);

    while(!monitor.done)
    {
	globus_cond_wait(&monitor.cond, &monitor.mutex);
	if(globus_l_globusrun_ctrlc && (!globus_l_globusrun_ctrlc_handled))
	{
	    if(verbose)
	    {
		printf("Cancelling job...\n");
                globus_l_globusrun_remove_cancel_poll();
	    }
	    globus_gram_client_job_cancel(job_contact);
	    globus_l_globusrun_ctrlc_handled = GLOBUS_TRUE;
	}
    }
    globus_mutex_unlock(&monitor.mutex);

    unlink(jobfilename);

    globus_gram_client_callback_disallow(callback_contact);
    globus_free(callback_contact);

    globus_mutex_destroy(&monitor.mutex);
    globus_cond_destroy(&monitor.cond);
    
    err=!(monitor.job_state == GLOBUS_GRAM_CLIENT_JOB_STATE_DONE);

hard_exit:

    if(job_contact != GLOBUS_NULL)
    {
	globus_gram_client_job_contact_free(job_contact);
    }
    if(jobfilename != GLOBUS_NULL)
    {
	globus_free(jobfilename);
    }
    globus_module_deactivate(GLOBUS_GRAM_CLIENT_MODULE);

    return err;
} /* globus_l_globusrun_gramrun() */

/******************************************************************************
Function: globus_l_globusrun_durocrun()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static int
globus_l_globusrun_durocrun(char *request_string,
			    unsigned long options,
			    int mpi_version)
{
    globus_duroc_control_t control;
    char *job_contact = GLOBUS_NULL;
    int results_count;
    int *results;
    int err=0;
    char *callback_contact;
    char *jobfilename = GLOBUS_NULL;

    
    globus_bool_t verbose = !(options & (GLOBUSRUN_ARG_QUIET));

    /* trap SIGINTs until job is submitted, then potentially ignore them */
    globus_l_globusrun_signal(SIGINT,
                              globus_l_globusrun_sigint_handler);
#   if defined(BUILD_LITE)
    {
	globus_poll_add(globus_l_globusrun_signal_wakeup, 1);
    }
#   endif
    
    err = globus_duroc_control_init (&control);

    if(err != GLOBUS_SUCCESS)
    {
	globus_libc_fprintf(stderr,
			    "Error initializing duroc control (%d)\n",
			    err);
	goto user_exit;
    }

    if(verbose)
    {
	globus_libc_fprintf(stderr,
			   "making globus_duroc request: %s\n",
			   request_string);
    }
    
    err = globus_duroc_control_job_request (&control,
					    request_string,
					    0, 
					    GLOBUS_NULL, 
					    &job_contact,
					    &results_count,
					    &results);

    if(verbose)
    {
	globus_libc_printf("duroc request status: %d\n"
			   "duroc job contact: \"%s\"\n",
			   err,
			   ((err==GLOBUS_DUROC_SUCCESS)
			    ? job_contact
			    : ""));
    }

    if(globus_l_globusrun_ctrlc && (!globus_l_globusrun_ctrlc_handled))
    {
        if(verbose)
        {
            globus_libc_printf("Cancelling job\n");
            globus_l_globusrun_remove_cancel_poll();
        }

        globus_duroc_control_job_cancel(&control,
                                        job_contact);
        globus_l_globusrun_ctrlc_handled = GLOBUS_TRUE;
        goto user_exit;
    }

    if(options & GLOBUSRUN_ARG_IGNORE_CTRLC)
    {
	globus_l_globusrun_signal(SIGINT,
                                  SIG_DFL);
#       if defined(BUILD_LITE)
	{
	    globus_poll_remove(globus_l_globusrun_signal_wakeup);
	}
#       endif
    }

    /* handle result of job request now */
    if ( err == GLOBUS_DUROC_SUCCESS )
    {
	int i;
	FILE *fp;
	char **subjob_labels;

	if(options & GLOBUSRUN_ARG_DRYRUN)
	{
	    globus_bool_t dryrun_ok=GLOBUS_TRUE;
	    if(verbose)
	    {
		char **subjob_labels;
		
		subjob_labels = globus_l_globusrun_duroc_subjob_labels(request_string);
		for(i = 0; i < results_count; i++)
		{
		    if(results[i] !=
		       GLOBUS_DUROC_SUCCESS)
		    {
			if(globus_duroc_error_is_gram_client_error(results[i]) &&
			   globus_duroc_error_get_gram_client_error(results[i]) != GLOBUS_GRAM_CLIENT_ERROR_DRYRUN)
			{
			    dryrun_ok = GLOBUS_FALSE;
			    globus_libc_printf("Duroc subjob (label = \"%s\") failed because %s (error code %d)\n",
					       subjob_labels[i],
					       globus_duroc_error_string(results[i]),
					       results[i]);
			    globus_free(subjob_labels[i]);
			}
		    }
		}
		if(dryrun_ok)
		{
		    globus_libc_printf("Dryrun successful\n");
		}
		globus_free(subjob_labels);
	    }
	    
	    goto user_exit;
	}
	else
	{
	    if(verbose)
	    {
		char **subjob_labels;
		
		subjob_labels = globus_l_globusrun_duroc_subjob_labels(request_string);

		globus_libc_printf("duroc subjob status:\n");
		for (i=0; i<results_count; i++)
		{
		    if(results[i] == GLOBUS_DUROC_SUCCESS)
		    {
			globus_libc_printf("    Submission of subjob (label = \"%s\") succeeded\n",
					   subjob_labels[i]);
		    }
		    else
		    {
			globus_libc_printf("    Submission of subjob (label = \"%s\") failed because %s (error code %d)\n",
					   subjob_labels[i],
					   globus_duroc_error_string(results[i]),
					   results[i]);
		    }
		    globus_free(subjob_labels[i]);
		}
		globus_free(subjob_labels);
	    }
	    
	    globus_free (results);
	    jobfilename = globus_l_globusrun_create_job_file();
	    if(jobfilename == GLOBUS_NULL)
	    {
		if(verbose)
		{
		    globus_libc_fprintf(stderr,
					"Error opening job state file\n");
		    
		}
	    }
	    else
	    {
		globus_l_globusrun_update_duroc_job_file(&control,
							 job_contact,
							 request_string,
							 jobfilename);
	    }
	}
    }
    else
    {
	if(options & GLOBUSRUN_ARG_DRYRUN)
	{
	    if(verbose)
	    {
		globus_libc_printf("Dryrun failure\n");
	    }
	}
	else
	{
	    if(verbose)
	    {
		globus_libc_fprintf(stderr,
				    "duroc request failed, exiting.\n");
	    }
	}
	return err;
    }

    if (options & GLOBUSRUN_ARG_INTERACTIVE)
    {
	
	globus_libc_printf ("entering interactive control mode...\n\n");
	
	/* loop for commands */
	while (1)
	{
	    int i;
	    int     subjob_count;
	    char ** subjob_labels;
	    int   * subjob_states;

            if(globus_l_globusrun_ctrlc && (!globus_l_globusrun_ctrlc_handled))
	    {
		if(verbose)
		{
		    globus_libc_printf("Cancelling job\n");
                    globus_l_globusrun_remove_cancel_poll();
		}

		globus_duroc_control_job_cancel(&control,
						job_contact);
                globus_l_globusrun_ctrlc_handled = GLOBUS_TRUE;
		goto user_exit;
	    }
	    /* print job state summary */
	    err = globus_duroc_control_subjob_states (&control,
						      job_contact,
						      &subjob_count,
						      &subjob_states,
						      &subjob_labels);
	    if(err != GLOBUS_SUCCESS)
	    {
		printf("Error polling duroc control subjob states (%d)\n",
		       err);
		goto user_exit;
	    }

	    globus_libc_fprintf (stdout, "subjob states:\n");
	    
	    for (i=0; i<subjob_count; i++)
	    {
		globus_libc_printf (
		    "subjob >>%s<< %s\n",
		    (subjob_labels[i] ? subjob_labels[i] : "(none)"),
		    ((subjob_states[i]
		      ==GLOBUS_DUROC_SUBJOB_STATE_PENDING)
		     ? "PEND"
		     : ((subjob_states[i]
			 ==GLOBUS_DUROC_SUBJOB_STATE_ACTIVE)
			? "pend ACTIVE"
			: ((subjob_states[i]
			    ==GLOBUS_DUROC_SUBJOB_STATE_CHECKED_IN)
			   ? "pend active CHECKIN"
			   : ((subjob_states[i]
			       ==GLOBUS_DUROC_SUBJOB_STATE_RELEASED)
			      ? "pend active checkin RUN"
			      : (subjob_states[i]
				 ==GLOBUS_DUROC_SUBJOB_STATE_DONE)
			      ? "DONE"
			      : "FAILED")))));
		globus_free (subjob_labels[i]);
		subjob_labels[i] = NULL;
	    }
	    
	    globus_libc_fprintf (stdout, "end subjob states.\n\n");
	    
	    globus_free (subjob_states);
	    globus_free (subjob_labels);
	    
	    /* prompt for user command */
	    
	    globus_libc_fprintf (stdout, "\n"
			   "enter command \"Dlabel\" (delete labeled subjob),\n"
			   "              \"K\" (kill entire job),\n"
			   "              \"C\" (commit current job),\n"
			   "           or \"Q\" (quit request tool)\n\n");
	    
	    /* get user command and perform background processing */
	    {
		globus_fifo_t input;
		
		err = globus_fifo_init (&input); 
		if(err != GLOBUS_SUCCESS)
		{
		    globus_libc_printf("Internal error intializing data structure\n");
		    goto user_exit;
		}
		
		while ( globus_fifo_empty (&input) ) {
		    ssize_t size;
		    char buf[1];
		    
                    if(globus_l_globusrun_ctrlc && (!globus_l_globusrun_ctrlc_handled))
		    {
			if(verbose)
			{
			    globus_libc_printf("Cancelling job\n");
                            globus_l_globusrun_remove_cancel_poll();
			}

			globus_duroc_control_job_cancel(&control,
							job_contact);
                        globus_l_globusrun_ctrlc_handled = GLOBUS_TRUE;
			goto user_exit;
		    }
		    size = read(fileno(stdin), buf, 1);
		    
		    if (size == 1)
		    {
			/* queue up data just read */
			globus_fifo_enqueue (&input, (void *) (long) buf[0]);
		    }
		    else if (size < 0)
		    {
			/* no input ready */
			globus_poll();
		    }
		    else
		    {
			/* eof? */
			globus_libc_fprintf(stdout, "eof reached. exiting.\n");
			
			goto user_exit;
		    }
		}
		
		if ( ((char) (long) globus_fifo_peek (&input)) == 'C' )
		{
		    globus_libc_fprintf (stdout, "C commit requested\n");
		    globus_fifo_dequeue (&input);
		    
		    globus_libc_fprintf (stdout,
					 "releasing barrier at "
					 "user's request...\n");
		    
		    err = globus_duroc_control_barrier_release (&control,
								job_contact,
								GLOBUS_TRUE);
		    
		    globus_libc_fprintf (stdout,
					 "release returned %s.\n",
					 (err ? "failure" : "success"));
		}
		else if ( ((char) (long) globus_fifo_peek (&input)) == 'Q' )
		{
		    globus_libc_fprintf (stdout, "Q quit requested\n");
		    globus_fifo_dequeue (&input);
		    
		    goto user_exit;
		}
		else if ( ((char) (long) globus_fifo_peek (&input)) == 'K' )
		{
		    globus_libc_fprintf (stdout, "K kill job requested\n");
		    globus_fifo_dequeue (&input);
		    
		    globus_libc_fprintf (stdout,
					 "canceling job at user's request.\n");
		    
		    err = globus_duroc_control_job_cancel (&control,
							   job_contact);
		    
		    globus_libc_fprintf (stdout,
					 "cancel returned %s.\n",
					 (err ? "failure" : "success"));
		}
		else if ( ((char) (long) globus_fifo_peek (&input)) == 'D' )
		{
		    /* get subjob label..
		     * all characters up to but not including newline */
		    int newline_read = 0;
		    ssize_t size;
		    char buf[1];
		    char *subjob_label;
		    
		    globus_fifo_dequeue (&input); /* throw out 'D' */
		    
		    while ( ! newline_read )
		    {
                        if(globus_l_globusrun_ctrlc && (!globus_l_globusrun_ctrlc_handled))
			{
			    if(verbose)
			    {
				printf("Cancelling job\n");
                                globus_l_globusrun_remove_cancel_poll();
			    }

			    globus_duroc_control_job_cancel(&control,
							    job_contact);
                            globus_l_globusrun_ctrlc_handled = GLOBUS_TRUE;
			    goto user_exit;
			}
			size = read (fileno(stdin), buf, 1);
			
			if (size == 1)
			{
			    if ( buf[0] != '\n' )
			    {
				globus_fifo_enqueue (&input,
						     (void *) (long) buf[0]);
			    }
			    else
			    {
				newline_read = 1;
			    }
			}
			else if ( size == -1 )
			{
			    /* no input ready */
			}
			else
			{
			    /* eof? */
			    globus_libc_fprintf (stdout,
						 "eof reached. exiting.\n");
			    goto user_exit;
			}
		    }
		    
		    {
			int len;
			
			len = globus_fifo_size (&input);
			if (len>0)
			{
			    subjob_label = globus_malloc (sizeof(char)
							  * (len + 1));
			    for (i=0; i<len; i++)
			    {
				subjob_label[i] = (char)
				    (long) globus_fifo_dequeue (&input);
			    }
			    subjob_label[len] = '\0';
			}
			else
			{
			    subjob_label = "";
			}
		    }
	    
		    globus_libc_fprintf (stdout,
					 "D delete subjob >>%s<< requested\n",
					 subjob_label);
		    
		    err = globus_duroc_control_subjob_delete (&control,
							      job_contact,
							      subjob_label);
		    
		    if (!err)
		    {
			globus_libc_fprintf (stdout,
					     "subjob >>%s<< deleted\n",
					     subjob_label);
		    }
		    else
		    {
			globus_libc_fprintf (stdout,
					     "subjob >>%s<< deletion failed "
					     "(code %d)\n",
					     subjob_label,
					     err);
		    }
		}
		else
		{
		    /* unknown character, reissue prompt */
		    globus_fifo_dequeue (&input);
		}
	    }
	    globus_l_globusrun_update_duroc_job_file(&control,
						     job_contact,
						     request_string,
						     jobfilename);
	    
	    globus_poll ();
	}
	
    after_loop2: /* eliminate pesky dead-code compiler warning */
	;
    }
    else
    {
	if(verbose)
	{
	    globus_libc_printf("releasing barrier in automatic mode...\n");
	}
	
	globus_duroc_control_barrier_release (&control,
					      job_contact,
					      GLOBUS_TRUE);
	
	if(verbose)
	{
	    globus_libc_printf("waiting for job termination\n");
	}
	
	while (GLOBUS_TRUE)
	{
	    int i;
	    int     subjob_count;
	    char ** subjob_labels;
	    int   * subjob_states;
	    int not_terminated = 0;
	    
	    /* poll for job state */	    
            if(globus_l_globusrun_ctrlc && (!globus_l_globusrun_ctrlc_handled))
	    {
		if(verbose)
		{
		    globus_libc_printf("Cancelling job\n");
                    globus_l_globusrun_remove_cancel_poll();
		}
		globus_duroc_control_job_cancel(&control,
						job_contact);
                globus_l_globusrun_ctrlc_handled = GLOBUS_TRUE;
		goto user_exit;
	    }
	    err = globus_duroc_control_subjob_states (&control,
						      job_contact,
						      &subjob_count,
						      &subjob_states,
						      &subjob_labels);
            if(err != GLOBUS_SUCCESS)
            {
                globus_libc_printf("Error polling duroc control subjob states (%d)\n",
                                   err);
                goto user_exit;
            }
 
	    for (i=0; i<subjob_count; i++)
	    {
		if ( (subjob_states[i] != GLOBUS_DUROC_SUBJOB_STATE_DONE) &&
		     (subjob_states[i] != GLOBUS_DUROC_SUBJOB_STATE_FAILED) )
		{
		    not_terminated = 1;
		}
		
		globus_free (subjob_labels[i]);
		subjob_labels[i] = NULL;
	    }
	    
	    globus_free (subjob_states);
	    globus_free (subjob_labels);
	    
	    if ( not_terminated )
	    {
		globus_l_globusrun_update_duroc_job_file(&control,
							 job_contact,
							 request_string,
							 jobfilename);
	    
		globus_poll_blocking ();
	    }
	    else
	    {
		goto job_terminated;
	    }
	}
	
    job_terminated:
	goto user_exit;
    }
    
user_exit:
    if(jobfilename != GLOBUS_NULL)
    {
	unlink(jobfilename);
    }
    globus_free(jobfilename);
    return 0;
} /* globus_l_globusrun_durocrun() */

/******************************************************************************
Function: globus_l_globusrun_fault_callback()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static int
globus_l_globusrun_fault_callback (void *user_arg, int fault_code)
{
    int debug=0;
    if(debug)
    {
	
	globus_libc_printf("globusrun received nexus fault code %d\n",
			   fault_code);
    }
    return 0;
} /* globus_l_globusrun_fault_callback() */

/******************************************************************************
Function: globus_l_globusrun_get_rm_contact()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static char *
globus_l_globusrun_get_rm_contact(char *resource)
{
    LDAP *ldap_server;
    int port=atoi(GLOBUS_MDS_PORT);
    char *base_dn=GLOBUS_MDS_ROOT_DN;
    char *search_string;
    char *server = GLOBUS_MDS_HOST;
    LDAPMessage *reply;
    LDAPMessage *entry;
    char *attrs[3];
    char *search_format=
	"(&(objectclass=GlobusResourceManager)"
	  "(|(mn=%s)(cn=%s)))";
    attrs[0] = "contact";
    attrs[1] = GLOBUS_NULL;
    
    if(strchr(resource, (int) ':') != GLOBUS_NULL)
    {
	return strdup(resource);
    }
	
    if((ldap_server = ldap_open(server, port)) == GLOBUS_NULL)
    {
	ldap_perror(ldap_server, "ldap_open");
	exit(1);
    }

    if(ldap_simple_bind_s(ldap_server, "", "") != LDAP_SUCCESS)
    {
	ldap_perror(ldap_server, "ldap_simple_bind_s");
	ldap_unbind(ldap_server);
	exit(1);
    }

    search_string=globus_malloc((2*strlen(resource))+
				strlen(search_format)+
				1);

    sprintf(search_string, search_format, resource, resource);
    
    if(ldap_search_s(ldap_server,
		     base_dn,
		     LDAP_SCOPE_SUBTREE,
		     search_string,
		     attrs,
		     0,
		     &reply) != LDAP_SUCCESS)
    {
	ldap_perror(ldap_server, "ldap_search");
	ldap_unbind(ldap_server);
	exit(1);
    }

    for(entry = ldap_first_entry(ldap_server, reply);
	entry != GLOBUS_NULL;
	entry = ldap_next_entry(ldap_server, entry))
    {
	char *contact;
	contact = globus_l_globusrun_get_contact_string(ldap_server, entry);
	if(contact != GLOBUS_NULL)
	{
	    ldap_unbind(ldap_server);
	    return contact;
	}
    }
    ldap_unbind(ldap_server);
    return GLOBUS_NULL;
} /* globus_l_globusrun_get_rm_contact() */

/******************************************************************************
Function: globus_l_globusrun_get_contact_string()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static char *
globus_l_globusrun_get_contact_string(LDAP *ldap_server,
				      LDAPMessage* entry)
{
    char *a, *dn;
    BerElement *ber;
    char** values;
    int numValues;
    int i;
    char *contact=GLOBUS_NULL;

    for (a = ldap_first_attribute(ldap_server,entry,&ber); a != NULL;
	 a = ldap_next_attribute(ldap_server,entry,ber) )
    {
	values = ldap_get_values(ldap_server,entry,a);
	numValues = ldap_count_values(values);
	
	if(strcmp(a, "contact") == 0)
	{
	    contact = strdup(values[0]);
	    ldap_value_free(values);
	    break;
	}
    }
    return contact;
} /* globus_l_globusrun_get_contact_string() */

/******************************************************************************
Function: globus_l_globusrun_create_globusrundir()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static char *
globus_l_globusrun_create_globusrundir(void)
{
    char *home, *globusdir;
    struct passwd pw, *pwptr;
    char pwbuf[1024];
    int rc;
    
    /* Find the globus jobs (~/.globus/globusrun) directory,
       creating it if it doesn't exist */
    rc = globus_libc_getpwuid_r(getuid(),
				&pw,
				pwbuf,
				1024,
				&pwptr);
    
    home = pwptr->pw_dir;

    globusdir=(char *) globus_malloc(PATH_MAX);

    if(globusdir == GLOBUS_NULL)
    {
	return globusdir;
    }

    globus_libc_sprintf(globusdir,
			"%s/.globus",
			home);
    rc = mkdir(globusdir, 0777);
    if(rc != GLOBUS_SUCCESS && errno != EEXIST)
    {
	globus_libc_printf("Error creating %s\n",
			   globusdir);
	globus_free(globusdir);
	return GLOBUS_NULL;
    }

    if(strcat(globusdir, "/globusrun") == GLOBUS_NULL)
    {
	globus_free(globusdir);
	return GLOBUS_NULL;
    }

    rc = mkdir(globusdir, 0777);
    if(rc != GLOBUS_SUCCESS && errno != EEXIST)
    {
	globus_libc_printf("Error creating %s\n",
			   globusdir);
	globus_free(globusdir);
	return GLOBUS_NULL;
    }

    return globusdir;
}

/******************************************************************************
Function: globus_l_globusrun_create_job_file()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static char *
globus_l_globusrun_create_job_file(void)
{
    char *globusdir, *tmp_filename, *real_filename;
    int rc;
    globus_bool_t done = GLOBUS_FALSE;
    int seed;
    char hostname[MAXHOSTNAMELEN];
    
    /* Find the globus jobs (~/.globus/globusrun) directory,
       creating it if it doesn't exist */
    globusdir = globus_l_globusrun_create_globusrundir();
    if(globusdir == GLOBUS_NULL)
    {
	return GLOBUS_NULL;
    }
    
    tmp_filename = (char *) globus_malloc(PATH_MAX);
    if(tmp_filename == GLOBUS_NULL)
    {
	globus_free(globusdir);
	return GLOBUS_NULL;
    }
    
    globus_libc_gethostname(hostname, MAXHOSTNAMELEN);
    
    globus_libc_sprintf(tmp_filename,
			"%s/%s-%ld-%ld",
			globusdir,
			hostname,
			(long) globus_libc_getpid(),
			(long) globus_thread_self());

    globus_libc_lock();
    rc = creat(tmp_filename, 0777);
    globus_libc_unlock();

    if(rc < 0)
    {
	globus_free(tmp_filename);
	globus_free(globusdir);
	return GLOBUS_NULL;
    }
    globus_libc_close(rc);
    real_filename = (char *) globus_malloc(PATH_MAX);
    if(real_filename == GLOBUS_NULL)
    {
	globus_free(globusdir);
	unlink(tmp_filename);
	globus_free(tmp_filename);
	return GLOBUS_NULL;
    }
    seed = ((int) (long) time(GLOBUS_NULL)) % 32000;
    if(seed == 0)
    {
	seed++;
    }

    while(!done)
    {
	struct stat gstat;
	
	globus_libc_sprintf(real_filename,
			    "%s/%05d",
			    globusdir,
			    seed);

	seed = (seed + 1) % 32000;
	if(seed == 0)
	{
	    seed++;
	}
	    
	link(tmp_filename,
	     real_filename);

	while(stat(tmp_filename, &gstat));
	if(gstat.st_nlink == 2)
	{
	    unlink(tmp_filename);
	    done = GLOBUS_TRUE;
	}
    }

    globus_free(tmp_filename);
    globus_free(globusdir);
    return real_filename;
}

/******************************************************************************
Function: globus_l_globusrun_list_jobs()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static void
globus_l_globusrun_list_jobs(void)
{
    char *globusdirname;
    DIR *globusdir;
    struct dirent *dir_entry;
    globus_bool_t first = GLOBUS_TRUE;
    globus_byte_t file_version;
    
    globusdirname = globus_l_globusrun_create_globusrundir();
    if(globusdirname == GLOBUS_NULL)
    {
	return;
    }

    chdir(globusdirname);
    
    globusdir = opendir(globusdirname);
    globus_free(globusdirname);
    if(globusdir == GLOBUS_NULL)
    {
	return;
    }
    else
    {
	for(globus_libc_readdir_r(globusdir, &dir_entry);
	    dir_entry != GLOBUS_NULL;
	    globus_libc_readdir_r(globusdir, &dir_entry))
	{
	    int job_id;
	    
	    if(strcmp(dir_entry->d_name, ".") == 0 ||
	       strcmp(dir_entry->d_name, "..") == 0)
	    {
		globus_free(dir_entry);
		continue;
	    }
	    else if ((job_id = atoi(dir_entry->d_name)) == 0)
	    {
		globus_free(dir_entry);
		continue;
	    }
	    else
	    {
		FILE *fp;
		char *buf;
		int rc;
		int length;

		fp = fopen(dir_entry->d_name, "r");
		if(fscanf(fp,
			  "GR%c:",
			  &file_version) == 1)
		{
		    if(file_version != globus_l_globusrun_file_version)
		    {
			goto skip_file;
		    }
		}
		else
		{
		    goto skip_file;
		}
		if(first)
		{
		    globus_libc_printf("JobID   RSL\n");
		    first = GLOBUS_FALSE;
		}
		globus_libc_printf("%5d  ", job_id);
		   
		if(fscanf(fp,
			  "%d:",
			  &length))
		{
		    buf = globus_malloc(length+1);
		    fread(buf, sizeof(char), length, fp);
		    buf[length] = '\0';
		    globus_libc_printf("%s\n", buf);
		    globus_free(buf);
		}
		
	    skip_file:
		fclose(fp);
		globus_free(dir_entry);
	    }
	}
	if(first)
	{
	    globus_libc_printf("<no outstanding globusrun jobs>\n");
	}
    }
    globus_libc_closedir(globusdir);
}

/******************************************************************************
Function: globus_l_globusrun_kill_job()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static int
globus_l_globusrun_kill_job(int job_id,
			    unsigned long options)
{
    char *globusdirname;
    FILE *fp;
    int length;
    char *buf;
    char filename[10];
    globus_byte_t version;
    int err;
    
    globusdirname = globus_l_globusrun_create_globusrundir();
    if(globusdirname == GLOBUS_NULL)
    {
	return -1;
    }


    chdir(globusdirname);
    globus_free(globusdirname);

    globus_libc_lock();
    sprintf(filename, "%05d", job_id);
    fp = fopen(filename, "r");
    globus_libc_unlock();

    if(fp == GLOBUS_NULL)
    {
	globus_libc_printf("Invalid job id: %d\n", job_id);
	return -1;
    }

    err = globus_module_activate(GLOBUS_GRAM_CLIENT_MODULE);

    if ( err != GLOBUS_SUCCESS )
    {
	globus_libc_fprintf(stderr, "Error initializing GRAM_CLIENT\n");
	return -1;
    }

    globus_libc_lock();

    if(fscanf(fp,
	      "GR%c:",
	      &version) == 1)
    {
	if(version != globus_l_globusrun_file_version)
	{
	    globus_libc_printf("Incompatible job version: %d\n",
			       (int) version);
	    return -1;
	}
    }
    else
    {
	globus_libc_printf("Incompatible job version: <unknown>\n");
	return -1;
    }
    fscanf(fp,
	   "%d:",
	   &length);
    fseek(fp,
	  length,
	  SEEK_CUR);

    while(fscanf(fp,
		 "%d:",
		 &length) != EOF)
    {
	buf = malloc(length+1);
	fread(buf, sizeof(char), length, fp);
	buf[length] = '\0';

	globus_libc_unlock();
	
	globus_gram_client_job_cancel(buf);

	globus_libc_lock();

	free(buf);
    }

    
    if ( !(options & GLOBUSRUN_ARG_BATCH) )
    {
	unlink(filename);
    }
    
    globus_libc_unlock();
    
    globus_module_deactivate(GLOBUS_GRAM_CLIENT_MODULE);
    
    return GLOBUS_TRUE;
}

/******************************************************************************
Function: globus_l_globusrun_sigint_handler()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static RETSIGTYPE
globus_l_globusrun_sigint_handler(int dummy)
{
    globus_l_globusrun_ctrlc = GLOBUS_TRUE;

    /* don't trap any more signals */
    globus_l_globusrun_signal(SIGINT, SIG_DFL);
} /* globus_l_globusrun_sigint_handler() */

#if defined(BUILD_LITE)
/******************************************************************************
Function: globus_l_globusrun_signal_wakeup()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static int
globus_l_globusrun_signal_wakeup(globus_bool_t blocking_allowed)
{
    static int foo=0;

    if(foo < 100)
    {
	foo++;
    }
    else
    {
	globus_libc_usleep(10000);
	foo=0;
    }

    return globus_l_globusrun_ctrlc;
} /* globus_l_globusrun_signal_wakeup() */
#endif

/******************************************************************************
Function: globus_l_globusrun_update_duroc_job_file()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static void
globus_l_globusrun_update_duroc_job_file(globus_duroc_control_t *control,
					 char *job_contact,
					 char *request_string,
					 char *jobfilename)
{
    FILE *fp;
    int i, err;
    int results_count;
    char **subjob_contacts, **subjob_labels;
    
    globus_libc_lock();
    truncate(jobfilename, 0);
    fp = fopen(jobfilename, "w");
    globus_libc_unlock();

    if(fp)
    {
	globus_libc_fprintf(fp,
			    "GR%c:%d:%s",
			    globus_l_globusrun_file_version,
			    (int) strlen(request_string),
			    request_string);
	err = globus_duroc_control_subjob_contacts(control,
						   job_contact,
						   &results_count,
						   &subjob_contacts,
						   &subjob_labels);
	for(i = 0; i < results_count; i++)
	{
	    globus_libc_fprintf(fp,
				"%d:%s",
				(int) strlen(subjob_contacts[i]),
				subjob_contacts[i]);
	    globus_free(subjob_contacts[i]);
	    globus_free(subjob_labels[i]);
	}
	globus_free(subjob_labels);
	globus_free(subjob_contacts);
	
	fclose(fp);
    }
}

/******************************************************************************
Function: globus_l_globusrun_duroc_subjob_labels()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static char **
globus_l_globusrun_duroc_subjob_labels(char *request_string)
{
    char **subjob_labels;
    globus_rsl_t *ast;
    globus_list_t *subjob_list;
    globus_rsl_t *subjob;
    int list_size;
    int i;
    
    ast = globus_rsl_parse(request_string);

    subjob_list = globus_rsl_boolean_get_operand_list(ast);

    list_size = globus_list_size(subjob_list);

    subjob_labels = globus_malloc(list_size * sizeof(char *));
    
    for(i = 0; i < list_size; i++)
    {
	char **values;
	
	subjob = globus_list_first(subjob_list);

	globus_rsl_param_get(subjob,
			     GLOBUS_RSL_PARAM_SINGLE_LITERAL,
			     "label",
			     &values);

	globus_libc_lock();
	if(values[0] != GLOBUS_NULL)
	{
	    subjob_labels[i] = strdup(values[0]);
	}
	else
	{
	    subjob_labels[i] = strdup("<no label>");
	}
	globus_libc_unlock();

	globus_free(values);

	subjob_list = globus_list_rest(subjob_list);
    }

    globus_rsl_free_recursive(ast);

    return subjob_labels;
} /* globus_l_globusrun_duroc_subjob_labels() */

/******************************************************************************
Function: globus_l_globusrun_signal()
 
Description: 
 
Parameters: 
 
Returns: 
******************************************************************************/
static int
globus_l_globusrun_signal(int signum, RETSIGTYPE (*func)(int))
{
    struct sigaction act;

    memset(&act, '\0', sizeof(struct sigaction));
    sigemptyset(&(act.sa_mask));
    act.sa_handler = func;
    act.sa_flags = 0;

    return sigaction(signum, &act, GLOBUS_NULL);
} /* globus_l_globusrun_signal() */
