/******************************************************************************
CIF_Comm_Connection.H

Description:

  An abstract base class from which all other CIF_Comm connection objects are
  derived.  The class also includes the defintions of events and properties
  associated with a connection.
  
CVS Information:

  $Source: /home/dsl/Master/cif/CIF-Comm-C++-Current/library/CIF_Comm_Connection.H,v $
  $Date: 1998/10/09 01:00:31 $
  $Revision: 1.6 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

#if !defined(CIF_INCLUDE_COMM_CONNECTION)
#define CIF_INCLUDE_COMM_CONNECTION 1

/******************************************************************************
		       Define CIF_Comm_Connection class
******************************************************************************/
class
CIF_Comm_Connection
: public CIF_Toolkit_Object
{
    friend class CIF_Comm_Factory;
    
    /*
     * CIF_Comm_Connection::Message_Received_Event
     *
     * Event delivered to registered event listener when a new message is
     * received by the connection.
     */
    public:
    class Message_Received_Event
    : public CIF_Toolkit_Event
    {
	public:
	static const CIF_Toolkit_UUID		CLASS_ID;
	
	public:
	Message_Received_Event(
	    CIF_Comm_Connection *		connection);

	public:
	CIF_Comm_Connection *
	get_Connection() const;

	private:
	CIF_Comm_Connection *			l_connection;
    };

    /*
     * CIF_Comm_Connection::Properties
     *
     * Properties to be associated with a conection.  At present, properties
     * can only be specified when the connection is opened and remain in effect
     * until the connection is closed.
     *
     * At present, the only property is a pointer to an event listener which is
     * expecting to receive a Message_Received_Event when a new message
     * arrives.  See CIF_I_Comm_Properties.H for methods associated with the
     * base properties class.
     */
    public:
    class Properties
    : public CIF_I_Comm_Properties
    {
    };

    /*
     * CIF_Comm_Connection::open()
     */
    protected: virtual
    CIF_Bool
    open(
	ARG_IN const char *			url,
	ARG_IN const Properties &		properties,
	ARG_IN const CIF_Toolkit_Timeout &	timeout) = 0;

    /*
     * CIF_Comm_Connection::send()
     */
    public: virtual
    CIF_Bool
    send(
	ARG_IN const CIF_Byte *			buf,
	ARG_IN const CIF_Size			len) = 0;

    public: virtual
    CIF_Bool
    send(
	ARG_IN const CIF_Byte *			buf,
	ARG_IN const CIF_Size			len,
	ARG_IN const int			tag) = 0;

    /*
     * CIF_Comm_Connection::receive()
     */
    public: virtual
    CIF_Bool
    receive(
	ARG_OUT CIF_Comm_Message &		message) = 0;
    
    public: virtual
    CIF_Bool
    receive(
	ARG_OUT CIF_Comm_Message &		message,
	ARG_IN const CIF_Toolkit_Timeout &	timeout) = 0;
    
    /*
     * CIF_Comm_Connection::poll()
     */
    public: virtual
    CIF_Bool
    poll() = 0;

    /*
     * CIF_Comm_Connection::enable_Event_Listener()
     */
    public: virtual
    CIF_Bool
    enable_Event_Listener() = 0;

    /*
     * CIF_Comm_Connection::disable_Event_Listener()
     */
    public: virtual
    CIF_Bool
    disable_Event_Listener() = 0;

    /*
     * CIF_Comm_Connection::close()
     */
    public: virtual
    CIF_Bool
    close() = 0;
};

#endif /* CIF_INCLUDE_COMM_CONNECTION */
