/******************************************************************************
CIF_I_Comm_Connection.H

Description:

  CIF_Comm Connection class implementation

CVS Information:

  $Source: /home/dsl/Master/cif/CIF-Comm-C++-Current/library/CIF_I_Comm_Connection.H,v $
  $Date: 1998/10/09 01:00:32 $
  $Revision: 1.6 $
  $State: Exp $
  $Author: toonen $
******************************************************************************/

#if !defined(CIF_INCLUDE_I_COMM_CONNECTION)
#define CIF_INCLUDE_I_COMM_CONNECTION 1

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF_Comm.H"
#include "CIF_I_Comm_Endpoint.H"
#include "CIF_I_Comm_Startpoint.H"

#include "globus_nexus.h"
#include "globus_common.h"


/******************************************************************************
		       Define CIF_I_Comm_Connection class
******************************************************************************/

class
CIF_I_Comm_Connection
: public CIF_Comm_Connection,
  public CIF_Toolkit_Event_Listener
{
    friend CIF_Comm_Factory;
    
    /**************************************************************************
				    API methods
    **************************************************************************/

    /*
     * CIF_I_Comm_Connection::CIF_I_Comm_Connection()
     *
     */
    protected:
    CIF_I_Comm_Connection();

    /*
     * CIF_I_Comm_Connection::open()
     */
    protected:
    CIF_Bool
    open(
	ARG_IN const char *			url,
	ARG_IN const
	CIF_Comm_Connection::Properties &	properties,
	ARG_IN const CIF_Toolkit_Timeout &	timeout);
    
    protected:
    CIF_Bool
    open(
	ARG_INOUT globus_nexus_buffer_t &	nx_buf,
	ARG_IN const
	CIF_Comm_Connection::Properties &	properties,
	ARG_IN const CIF_Toolkit_Timeout &	timeout);
	
    /*
     * CIF_I_Comm_Connection::send()
     */
    protected: virtual
    CIF_Bool
    send(
	ARG_IN const CIF_Byte *			msg_buf,
	ARG_IN const CIF_Size			msg_len);

    protected: virtual
    CIF_Bool
    send(
	ARG_IN const CIF_Byte *			msg_buf,
	ARG_IN const CIF_Size			msg_len,
	ARG_IN const int			msg_tag);
    
    /*
     * CIF_I_Comm_Connection::receive()
     */
    protected: virtual
    CIF_Bool
    receive(
	ARG_OUT CIF_Comm_Message &		message);

    protected: virtual
    CIF_Bool
    receive(
	ARG_OUT CIF_Comm_Message &		message,
	ARG_IN const CIF_Toolkit_Timeout &	timeout);

    /*
     * CIF_I_Comm_Connection::poll()
     */
    protected: virtual
    CIF_Bool
    poll();

    /*
     * CIF_I_Comm_Connection::enable_Event_Listener()
     */
    public: virtual
    CIF_Bool
    enable_Event_Listener();

    /*
     * CIF_I_Comm_Connection::disable_Event_Listener()
     */
    public: virtual
    CIF_Bool
    disable_Event_Listener();

    /*
     * CIF_I_Comm_Connection::close()
     *
     * Close an established or faulty connection, freeing all resources,
     * including queued messages.
     */
    protected: virtual
    CIF_Bool
    close();

    /*
     * CIF_I_Comm_Connection::~CIF_I_Comm_Connection()
     *
     * Free any resources used by the connection.  If the connection is still
     * open, it will be closed.
     */
    protected: virtual
    ~CIF_I_Comm_Connection();


    /**************************************************************************
	   Internal data structures (some accessible by derived classes)
    **************************************************************************/

    public:
    enum Connection_Flow_Mask
    {
	CONNECTION_UNESTABLISHED = 0,
	CONNECTION_OUTGOING = 1,
	CONNECTION_INCOMING = 2,
	CONNECTION_BIDIRECTIONAL = 3
    };

    private:
    enum Object_State
    {
	OBJECT_STATE_INACTIVE,
	OBJECT_STATE_ACTIVE,
	OBJECT_STATE_DELETED
    };

    
    /**************************************************************************
		        Methods supplied by derived classes
    **************************************************************************/
    protected: virtual
    Connection_Flow_Mask
    open_Connection(
	ARG_IN const char *			url,
	ARG_INMOD
	CIF_I_Comm_Endpoint::Properties &	ep_properties,
	ARG_IN const CIF_Toolkit_Timeout &	timeout) = 0;

    protected: virtual
    Connection_Flow_Mask
    open_Connection(
	ARG_INOUT globus_nexus_buffer_t &	nx_buf,
	ARG_INMOD
	CIF_I_Comm_Endpoint::Properties &	ep_properties,
	ARG_IN const CIF_Toolkit_Timeout &	timeout) = 0;


    /**************************************************************************
			         Internal methods
    **************************************************************************/

    private:
    CIF_Bool
    l_send(
	ARG_IN const CIF_Byte *			msg_buf,
	ARG_IN const CIF_Size			msg_len,
	ARG_IN const int			msg_tag);
    
    private:
    CIF_Bool
    l_receive(
	ARG_OUT CIF_Comm_Message &		message,
	ARG_IN const CIF_Toolkit_Timeout &	timeout);

    private:
    CIF_Bool
    l_close(
	ARG_INMOD CIF_Toolkit_Monitor &		monitor);
    
    private:
    CIF_Comm_Status
    l_verify_Object_State(
	ARG_IN const Object_State &		desired_state);
    
    private:
    void
    handle_Event(
	ARG_INMOD CIF_Toolkit_Event &		ep_event);
    

    /**************************************************************************
	    Internal data members (some accessible by derived classes)
    **************************************************************************/

    /* mutex to protect listener data structures */
    protected:
    CIF_Toolkit_Mutex				l_mutex;

    /* startpoint and endpoint used to communicate data */
    protected:
    CIF_I_Comm_Endpoint				l_ep;
    CIF_I_Comm_Startpoint			l_sp;

    /* data flow(s) supported by the connection */
    private:
    Connection_Flow_Mask			l_flow;	

    /* event listener to notify when a connection request arrives */
    private:
    CIF_Toolkit_Event_Listener *		l_event_listener;

    /* synchronization variables to ensure that a close doesn't occur while
       other threads using the object are waiting for blocking operations to
       complete. */
    CIF_Toolkit_Cond				l_close_cond;
    volatile int				l_blocking_cnt;
    volatile CIF_Bool				l_close_pending;
    
    /* state of the connection object */
    volatile Object_State			l_object_state;

    /**************************************************************************
			     Internal (local) methods
    **************************************************************************/

};


#endif /* CIF_INCLUDE_I_COMM_CONNECTION */
